//#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* Dtapi.cpp *#*#*#*#*#*#*#*#*# (C) 2000-2004 DEKTEC
//
// Implementation of global DTAPI functions.

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- License -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// This program is free software; You can freely use/modify this source code in any way
// you desire, including for commercial applications.
//

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Change History -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//	TD	2007.02.08	Configuraqble IO support for DTA-105 added
//	TD	2007.01.31	Dta1xx IO configuration is read from the driver
//	TD	2007.01.29	Added support for DTA-160 ASI-only
//	TD	2007.01.23	Added support for the DTA-145 and DTA-2145
//	KvR	2006.08.14	Added support for DTU-245 + some updates for DTA-105
//	JH	2005.07.13	Added support for DTA-105/110/545
//	MG	2004.11.23	Added support for Dtu2xx driver
//	MG	2004.10.20	Adapted for new DtDevice class
//	MG	19-11-2002	Created	based on DTAPI for windows

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Include files -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
#include "StdAfx.h"

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Constants -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

// Minimum version of Dta1xx device driver with which DTAPI can work
const int  Dta1xxMinMajor	= 1;
const int  Dta1xxMinMinor	= 11;
const int  Dta1xxMinBugFix	= 0;

// Minimum version of Dtu2xx device driver with which DTAPI can work
const int  Dtu2xxMinMajor	= 2;
const int  Dtu2xxMinMinor	= 0;
const int  Dtu2xxMinBugFix	= 1;

// Maximum supported number of PCI and USB devices
const int  MaxPciDevices	= 50;
const int  MaxUsbDevices	= 25;
const int  MaxDevices		= MaxPciDevices+MaxUsbDevices;

// Device driver prefixes
const char c_Dta1xxPrefix[] = "/dev/Dta1xx";
const char c_Dtu2xxPrefix[] = "/dev/usb/dektec/dtu2xx-";

//.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtapiCheckDeviceDriverVersion -.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Check whether the device driver versions are compatible with this version of the DtAPI.
// This overloaded variant of DtapiCheckDeviceDriverVersion reads the version of the
// installed drivers.
//
// Return value:
//	DTAPI_OK				Driver compatible with DtAPI
//	DTAPI_E_DRIVER_INCOMP	Driver not compatible with DtAPI
//	DTAPI_E_NO_DEVICE		No DEKTEC device in system: can't determine compatibility
//
DTAPI_RESULT  DtapiCheckDeviceDriverVersion(void)
{
	int  PciVersionMajor, PciVersionMinor, PciVersionBugFix, PciVersionBuild;
	int  UsbVersionMajor, UsbVersionMinor, UsbVersionBugFix, UsbVersionBuild;

	// Read Dta1xx and Dtu2xx device-driver version
	DTAPI_RESULT  DrPci = ::DtapiGetDeviceDriverVersion(DTAPI_CAT_PCI,
					PciVersionMajor, PciVersionMinor, PciVersionBugFix, PciVersionBuild);
	DTAPI_RESULT  DrUsb = ::DtapiGetDeviceDriverVersion(DTAPI_CAT_USB,
					UsbVersionMajor, UsbVersionMinor, UsbVersionBugFix, UsbVersionBuild);

	// If no DEKTEC device is found => we can't know whether driver is compatible
	if (DrPci==DTAPI_E_NO_DEVICE && DrUsb==DTAPI_E_NO_DEVICE)
		return DTAPI_E_NO_DEVICE;

	if (DrPci != DTAPI_E_NO_DEVICE)  {
		// We have been able to read the DTA-1xx device-driver version.
		// Compare with the minimum.
		DrPci =  ::DtapiCheckDeviceDriverVersion(DTAPI_CAT_PCI,
					PciVersionMajor, PciVersionMinor, PciVersionBugFix, PciVersionBuild);
	}
	if (DrUsb != DTAPI_E_NO_DEVICE)  {
		// We have been able to read a DTU-2xx device-driver version.
		// Compare with the minimum.
		DrUsb =  ::DtapiCheckDeviceDriverVersion( DTAPI_CAT_USB,
					UsbVersionMajor, UsbVersionMinor, UsbVersionBugFix, UsbVersionBuild);
	}

	// If either Dta2xx or the Dtu2xx driver is incompatible => return error
	// However if one is missing and the other complies => return DTAPI_OK
	//  (e.g. DrPci==DTAPI_E_NO_DEVICE and DrUsb==DTAPI_OK)
	if (DrPci==DTAPI_E_DRIVER_INCOMP || DrUsb==DTAPI_E_DRIVER_INCOMP)
		return DTAPI_E_DRIVER_INCOMP;

	return DTAPI_OK;
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtapiDeviceScan -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Scan for the devices available in the current system.
// The maximum number of devices this function can return is 10 PCI and 10 USB devices.
//
DTAPI_RESULT  DtapiDeviceScan(
	int  NumEntries,			// Number of function entries in pHwFuncs
	int&  NumEntriesResult,		// #Functions found or #entries required in pHwFuncs
	DtDeviceDesc*  pDvcDesc)	// Device descriptors for devices found
{
	// Let's be cautious
	if (pDvcDesc == NULL)
		return DTAPI_E_BUF_TOO_SMALL;

	// Let generic ::DtapiDeviceScan do the job
	return ::DtapiDeviceScan(SCAN_FULL, NumEntries, NumEntriesResult, pDvcDesc,
							 NULL, 0, 0, 0, 0, 0);
}


//-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtapiGetDeviceDriverVersion -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Get device driver version.
// Included for backwards compatibility only.
// Replaced by overloaded variant with DvcCategory argument.
//
DTAPI_RESULT  DtapiGetDeviceDriverVersion(
	int&  DriverVersionMajor,
	int&  DriverVersionMinor,
	int&  DriverVersionBugFix,
	int&  DriverVersionBuild)
{
	// We need a handle to a DEKTEC card for getting the device driver version.
	// So we execute a device scan.
	char  DevicePath[32];
	for (int nDeviceIndex=0; nDeviceIndex<MaxPciDevices; nDeviceIndex++) {

		// Construct path and try to open device
		sprintf(DevicePath, "%s%d", c_Dta1xxPrefix, nDeviceIndex);
      	        int  fdDta1xx = open(DevicePath, O_RDONLY | O_WRONLY);
		if ( fdDta1xx < 0 ) {
			continue;
		}


		// Found a DTA-1xx PCI card. Get device-driver version.

  		// Get location info
		DTA1XX_IOCTL_DATA iod;
		memset(&iod, 0, sizeof(DTA1XX_IOCTL_DATA));
		int nResult = ioctl(fdDta1xx, DTA1XX_IOCTL_GET_DRIVER_VERSION,  &iod);
		close(fdDta1xx);		// Do not leak handles

		// If DeviceIoControl succeeds => copy device-driver version and done.
		if (nResult==0 ) {
			DriverVersionMajor  = iod.m_Data.m_DriverVersion.m_nVersionMajor;
			DriverVersionMinor  = iod.m_Data.m_DriverVersion.m_nVersionMinor;
			DriverVersionBugFix = iod.m_Data.m_DriverVersion.m_nVersionMicro;
			DriverVersionBuild  = iod.m_Data.m_DriverVersion.m_nVersionBuild;

			return DTAPI_OK;
		}
	}

	// No DEKTEC card found. We can't return a device-driver version.
	return DTAPI_E_NO_DTA_CARD;
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtapiGetDeviceDriverVersion -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Get device driver version.
//
DTAPI_RESULT  DtapiGetDeviceDriverVersion(
	int	  DvcCat,				// Device category: DTAPI_CAT_PCI or DTAPI_CAT_USB
	int&  DriverVersionMajor,
	int&  DriverVersionMinor,
	int&  DriverVersionBugFix,
	int&  DriverVersionBuild)
{
	// Separate code for each device category
	if (DvcCat == DTAPI_CAT_PCI) {

		// We need a handle to a DEKTEC card for getting the device driver version.
		// So we execute a device scan.
		char  DevicePath[32];
		for (int nDeviceIndex=0; nDeviceIndex<MaxPciDevices; nDeviceIndex++) {

			// Construct path and try to open device
			sprintf(DevicePath, "%s%d", c_Dta1xxPrefix, nDeviceIndex);
			int  fdDta1xx = open(DevicePath, O_RDONLY | O_WRONLY );
			if ( fdDta1xx < 0 ) {
				continue;
			}

			// Found a DTA-1xx PCI card. Get device-driver version.

  			// Get version info
			DTA1XX_IOCTL_DATA iod;
			memset(&iod, 0, sizeof(DTA1XX_IOCTL_DATA));
			int nResult = ioctl(fdDta1xx, DTA1XX_IOCTL_GET_DRIVER_VERSION,  &iod);
			close(fdDta1xx);		// Do not leak handles

			// If DeviceIoControl succeeds => copy device-driver version and done.
			if (nResult==0 ) {
				DriverVersionMajor  = iod.m_Data.m_DriverVersion.m_nVersionMajor;
				DriverVersionMinor  = iod.m_Data.m_DriverVersion.m_nVersionMinor;
				DriverVersionBugFix = iod.m_Data.m_DriverVersion.m_nVersionMicro;
				DriverVersionBuild  = iod.m_Data.m_DriverVersion.m_nVersionBuild;

				return DTAPI_OK;
			}
		}
	} else {

		// We need a handle to a DEKTEC card for getting the device driver version.
		// So we execute a device scan.
		char  DevicePath[32];
		for (int DeviceIndex=0; DeviceIndex<MaxUsbDevices; DeviceIndex++) {

			// Construct path and try to open device
			sprintf(DevicePath, "%s%d", c_Dtu2xxPrefix, DeviceIndex);
			int  fdDtu2xx = open(DevicePath, O_RDONLY | O_WRONLY);
			if ( fdDtu2xx < 0 ) {
				continue;
			}

			// Found a DTU-2xx device. Get device-driver version.

			// Get version info
			DTU2XX_IOCTL_DATA iod;
			memset(&iod, 0, sizeof(DTA1XX_IOCTL_DATA));
			int nResult = ioctl(fdDtu2xx, DTU2XX_IOCTL_GET_DRIVER_VERSION,  &iod);
			close(fdDtu2xx);		// Do not leak handles

			// If DeviceIoControl succeeds => copy device-driver version and done.
			if (nResult==0 ) {
				DriverVersionMajor  = iod.m_Data.m_DriverVersion.m_nVersionMajor;
				DriverVersionMinor  = iod.m_Data.m_DriverVersion.m_nVersionMinor;
				DriverVersionBugFix = iod.m_Data.m_DriverVersion.m_nVersionMicro;
				DriverVersionBuild  = iod.m_Data.m_DriverVersion.m_nVersionBuild;

				return DTAPI_OK;
			}
		}

		// NOT supported yet
		return DTAPI_E_NO_DEVICE;
	}
	// No DEKTEC devices found. We can't return a device-driver version.
	return DTAPI_E_NO_DEVICE;
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtapiGetVersion -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Get version of DtAPI library.
//
DTAPI_RESULT  DtapiGetVersion(
	int&  LibVersionMajor,
	int&  LibVersionMinor)
{
	LibVersionMajor = DTAPI_VERSION_MAJOR;
	LibVersionMinor = DTAPI_VERSION_MINOR;

	return DTAPI_OK;
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtapiGetVersion -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Get version of DtAPI library, including bugfix and build number
//
DTAPI_RESULT  DtapiGetVersion(
	int&  LibVersionMajor,
	int&  LibVersionMinor,
	int&  LibVersionBugFix,
	int&  LibVersionBuild)
{
	LibVersionMajor = DTAPI_VERSION_MAJOR;
	LibVersionMinor = DTAPI_VERSION_MINOR;
	LibVersionBugFix = DTAPI_VERSION_BUGFIX;
	LibVersionBuild = DTAPI_VERSION_BUILD;

	return DTAPI_OK;
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtapiHwFuncScan -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Scan hardware functions available on the DTx-xxx devices in the current system.
//
DTAPI_RESULT  DtapiHwFuncScan(
	int  NumEntries,			// Number of function entries in pHwFuncs
	int&  NumEntriesResult,		// #Functions found or #entries required in pHwFuncs
	DtHwFuncDesc*  pHwFuncs)	// Hardware-function descriptions
{
	// Let's be cautious: do not crash if pointer argument is NULL
	if (pHwFuncs == NULL)
		return DTAPI_E_BUF_TOO_SMALL;

	// Scan devices
	int  NumDvcEntries = 0;
	DtDeviceDesc  DvcDesc[MaxDevices];
	DTAPI_RESULT  dr = DtapiDeviceScan(MaxDevices, NumDvcEntries, DvcDesc);
	if (dr == DTAPI_E_BUF_TOO_SMALL) {
		NumEntriesResult = MaxDevices;// Atelease MaxDevices entries are required
		return DTAPI_E_BUF_TOO_SMALL;
	}
	if (dr != DTAPI_OK)					// Should not be possible at the moment
		return dr;

	// Loop through devices
	DtHwFuncDesc*  pHwf = pHwFuncs;
	NumEntriesResult = 0;
	for (int DeviceIndex=0; DeviceIndex<NumDvcEntries; DeviceIndex++) {

		// Get type number
		int  TypeNumber = DvcDesc[DeviceIndex].m_TypeNumber;


		// Attach to the device
		DtDevice DtDvc;
		dr = DtDvc.AttachToSerial(DvcDesc[DeviceIndex].m_Serial);
		if ( dr != DTAPI_OK )
			continue;
		
		// Loop through hardware functions
		int InpIdx(0), OutpIdx(0);
		for (int Idx=0; Idx<DvcDesc[DeviceIndex].m_NumHwFuncs; Idx++) {

			// Copy to DtapiHwFunc structure, unless buffer is too small.
			NumEntriesResult++;
			if (NumEntries >= NumEntriesResult) {

				// Copy device descriptor
				pHwf->m_DvcDesc = DvcDesc[DeviceIndex];

				// Decode hardware-function type //TODOTD get from ioconfig
				pHwf->m_ChanType = 0;
				
				// Determine IO configaraion for USB devices
				if (  TypeNumber==205  || (TypeNumber==245 && Idx==1) )
				{
					pHwf->m_ChanType |= DTAPI_TS_OUTPUT;
				}
				else if ( TypeNumber==225 || (TypeNumber==245 && Idx==0) )
				{
					pHwf->m_ChanType |= DTAPI_TS_INPUT;
				}
				else  {
					// Get PCI board IO configuration
					int IoConfig = 0;
					dr = DtDvc.GetIoConfig(Idx+1, IoConfig);

					if (dr == DTAPI_OK) {
						if (IoConfig == DTAPI_IOCONFIG_INPUT) {
					pHwf->m_ChanType |= DTAPI_TS_INPUT;
						}
						else if (IoConfig == DTAPI_IOCONFIG_OUTPUT) {
							pHwf->m_ChanType |= DTAPI_TS_OUTPUT;
						}					
					}
				}

				// Flags not used yet
				pHwf->m_Flags = 0;

                // For now set port number sequentially
                pHwf->m_Port = (Idx+1);

				// Store the index (input or output)
				if ( pHwf->m_ChanType == DTAPI_TS_OUTPUT ) {
                                        pHwf->m_IndexOnDvc = OutpIdx;
					OutpIdx++;
				} else /*if ( pHwf->m_nHwFuncType == DTAPI_TS_INPUT )*/ {
					pHwf->m_IndexOnDvc = InpIdx;
					InpIdx++;
				}
				pHwf++;
			}
		}
		DtDvc.Detach();
	}
	return DTAPI_OK;
}

//+=+=+=+=+=+=+=+=+=+=+=+=+=+ DTAPI FUNCTIONS FOR INTERNAL USE +=+=+=+=+=+=+=+=+=+=+=+=+=+


//.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtapiCheckDeviceDriverVersion -.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Check whether the device driver version is compatible with this version of the DtAPI.
// This overloaded variant of DtapiCheckDeviceDriverVersion has the driver version as
// argument.
//
// Return value:
//	DTAPI_OK				Driver compatible with DtAPI
//	DTAPI_E_DRIVER_INCOMP	Driver not compatible with DtAPI
//
DTAPI_RESULT  DtapiCheckDeviceDriverVersion(
	int  DvcCat,			// Device category: DTAPI_CAT_PCI or DTAPI_CAT_USB
	int  VersionMajor,		// Driver version to be checked: major
	int  VersionMinor,		//   ,,      ,,   ,, ,,    ,,  : minor
	int  VersionBugFix,		//   ,,      ,,   ,, ,,    ,,  : bug fix
	int  VersionBuild)		//   ,,      ,,   ,, ,,    ,,  : build
{
	if (DvcCat == DTAPI_CAT_PCI) {
		if (VersionMajor < Dta1xxMinMajor)	 return DTAPI_E_DRIVER_INCOMP;
		if (VersionMajor > Dta1xxMinMajor)	 return DTAPI_OK;
		if (VersionMinor < Dta1xxMinMinor)	 return DTAPI_E_DRIVER_INCOMP;
		if (VersionMinor > Dta1xxMinMinor)	 return DTAPI_OK;
		if (VersionBugFix < Dta1xxMinBugFix) return DTAPI_E_DRIVER_INCOMP;
	}
	if (DvcCat == DTAPI_CAT_USB) {
		if (VersionMajor < Dtu2xxMinMajor)	 return DTAPI_E_DRIVER_INCOMP;
		if (VersionMajor > Dtu2xxMinMajor)	 return DTAPI_OK;
		if (VersionMinor < Dtu2xxMinMinor)	 return DTAPI_E_DRIVER_INCOMP;
		if (VersionMinor > Dtu2xxMinMinor)	 return DTAPI_OK;
		if (VersionBugFix < Dtu2xxMinBugFix) return DTAPI_E_DRIVER_INCOMP;
	}
	return DTAPI_OK;
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtapiDeviceScan -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Generic function for scanning/attaching DEKTEC devices.
// This function can return all devices, or attach to a specific device based on PCI
// slot, type number or serial number.
//
// The maximum number of devices this function can handle is 50 PCI and 25 USB devices.
//
DTAPI_RESULT  DtapiDeviceScan(
	int  ScanType,				// Type of scan. One of: SCAN_FULL, ATTACH_TO_SERIAL,
								// ATTACH_TO_SLOT, ATTACH_TO_TYPE
	// SCAN_FULL arguments:
	int  NumEntries,			// Number of device entries in pDvcDesc
	int&  NumEntriesResult,		// #Devices found or #entries required in pDvcDesc
	DtDeviceDesc*  DvcDescArr,	// Device descriptors for devices found.
								// Also used for ATTACH_TO_XXX scans
	// ATTACH_TO_XXX argument:
	DtDevice*  pDvc,			// Device structure
	// ATTACH_TO_SLOT arguments:
	int  PciBusNumber, int  SlotNumber,
	// ATTACH_TO_TYPE arguments:
	int  TypeNumber, int  DeviceNo,
	// ATTACH_TO_SERIAL argument:
	__int64  SerialNumber)		// Serial Number
{
	// Let's be cautious
	if (ScanType==SCAN_FULL && DvcDescArr==NULL)
		return DTAPI_E_INTERNAL;

	//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Device Scan -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

	// Initialise search variables
	DTAPI_RESULT dr;
	DtDeviceDesc  DvcDesc;
	DtDeviceDesc*  pDvcDesc = DvcDescArr;
	bool  DeviceFound = false;
	int  fdDriver;
	NumEntriesResult = 0;
	int  NumFound = 0;			// Incremented for every matching device (ATTACH_TO_TYPE)

	// First scan PCI devices (0<=Index<MaxPciDevices), than USB-2 devices
	// (MaxPciDevices<=Index<MaxDevices)
	for (int Index=0; Index<MaxDevices; Index++) {

		// PCI device or USB-2 device?
		int  Category = DTAPI_CAT_PCI;
		int  DeviceIndex = Index;
		const char*  PathPrefix = c_Dta1xxPrefix;
		if (Index >= MaxPciDevices) {
			if (ScanType == ATTACH_TO_SLOT)
				break;		// Don't scan USB devices when attaching to PCI Bus/Slot
			Category = DTAPI_CAT_USB;
			DeviceIndex -= MaxPciDevices;
			PathPrefix = c_Dtu2xxPrefix;
		}

		// Construct path and try to open device
		char  DevicePath[32];
		sprintf(DevicePath, "%s%d", PathPrefix, DeviceIndex);

		// First try read/write access
		fdDriver = open(DevicePath, O_RDWR );
		if ( fdDriver < 0 )
		{
			// Try write access only
			fdDriver = open(DevicePath, O_WRONLY );
			if ( fdDriver < 0 )
			{
				// Try read access only
				fdDriver = open(DevicePath, O_RDONLY );
				if ( fdDriver < 0 ) {
					continue;
				}
			}
		}
		
		// Found a new device. Get location info.
		DTA1XX_IOCTL_DATA iodDta1xx;
		DTU2XX_IOCTL_DATA iodDtu2xx;
		if (Category == DTAPI_CAT_PCI) {
			memset(&iodDta1xx, 0, sizeof(DTA1XX_IOCTL_DATA));
			dr = ioctl(fdDriver, DTA1XX_IOCTL_GET_LOCATION_INFO,  &iodDta1xx);
		}
		else if (Category == DTAPI_CAT_USB) {
			memset(&iodDtu2xx, 0, sizeof(DTU2XX_IOCTL_DATA));
			dr = ioctl(fdDriver, DTU2XX_IOCTL_GET_LOCATION_INFO,  &iodDtu2xx);
		}
		else
			dr = DTAPI_E; // Unknown type

		// If DeviceIoControl fails => forget about this device
		if (dr != DTAPI_OK) {
			close(fdDriver);		// Do not leak handles
			continue;
		}

		// Found a DEKTEC device!
		NumEntriesResult++;

		// Initialise a device descriptor
		DvcDesc.m_Category = Category;
		if (Category == DTAPI_CAT_PCI) {
			DvcDesc.m_PciBusNumber 	= iodDta1xx.m_Data.m_LocInfo.m_nBusNumber;
			DvcDesc.m_SlotNumber 	= iodDta1xx.m_Data.m_LocInfo.m_nSlotNumber;
			DvcDesc.m_UsbAddress 	= 0;
			DvcDesc.m_DeviceId 		= iodDta1xx.m_Data.m_LocInfo.m_nDeviceId;
			DvcDesc.m_VendorId 		= iodDta1xx.m_Data.m_LocInfo.m_nVendorId;
			DvcDesc.m_SubsystemId 	= iodDta1xx.m_Data.m_LocInfo.m_nSubSystemId;
			DvcDesc.m_SubVendorId 	= iodDta1xx.m_Data.m_LocInfo.m_nSubVendorId;
			DvcDesc.m_TypeNumber 	= DtapiDeviceId2TypeNumber((DvcDesc.m_VendorId == 0x10B5) ?
                                                       	DvcDesc.m_SubsystemId : DvcDesc.m_DeviceId);

			// Get firmware version
			DTA1XX_IOCTL_DATA iodFw;
			memset(&iodFw, 0, sizeof(DTA1XX_IOCTL_DATA));
			dr = ioctl(fdDriver, DTA1XX_IOCTL_GET_FW_VERSION, &iodFw);
			if (dr!=DTAPI_OK ) {
				close(fdDriver);
				return DTAPI_E_DEV_DRIVER;
			}
			DvcDesc.m_FirmwareVersion = iodFw.m_Data.m_uiValue;

		} else if ( Category == DTAPI_CAT_USB ) {

			DvcDesc.m_PciBusNumber 	= 0;
			DvcDesc.m_SlotNumber 	= 0;
			DvcDesc.m_UsbAddress 	= iodDtu2xx.m_Data.m_LocInfo.m_nUsbAddress;
			DvcDesc.m_DeviceId 		= iodDtu2xx.m_Data.m_LocInfo.m_nDeviceId;
			DvcDesc.m_VendorId 		= iodDtu2xx.m_Data.m_LocInfo.m_nVendorId;
			DvcDesc.m_SubsystemId 	= iodDtu2xx.m_Data.m_LocInfo.m_nSubSystemId;
			DvcDesc.m_SubVendorId 	= iodDtu2xx.m_Data.m_LocInfo.m_nSubVendorId;
			DvcDesc.m_TypeNumber 	= DtapiDeviceId2TypeNumber(
                                                         iodDtu2xx.m_Data.m_LocInfo.m_nSubSystemId);

			// Get firmware version
			DTU2XX_IOCTL_DATA iodFw;
			memset(&iodFw, 0, sizeof(DTU2XX_IOCTL_DATA));
			dr = ioctl(fdDriver, DTU2XX_IOCTL_GET_FW_VERSION, &iodFw);
			if (dr!=DTAPI_OK ) {
				close(fdDriver);
				return DTAPI_E_DEV_DRIVER;
			}
			DvcDesc.m_FirmwareVersion = iodFw.m_Data.m_uiValue;
		}

		// Derive number of hardware functions from type number
		Int NumFuncs = 0;
		Int NumInp = 0;
		Int NumOutp = 0;
		bool ConfigIo = false;
		switch (DvcDesc.m_TypeNumber) {
		case 100:  NumFuncs = 1; NumInp = 0; NumOutp = 1; break;
		case 102:  NumFuncs = 1; NumInp = 0; NumOutp = 1; break;
		case 105:  NumFuncs = 2; ConfigIo = true;					break;
		case 107:  NumFuncs = 1; NumInp = 0; NumOutp = 1; break;
		case 110:  NumFuncs = 1; NumInp = 0; NumOutp = 1; break;
		case 120:  NumFuncs = 1; NumInp = 1; NumOutp = 0; break;
		case 122:  NumFuncs = 1; NumInp = 1; NumOutp = 0; break;
		case 124:  NumFuncs = 4; NumInp = 4; NumOutp = 0; break;
		case 140:  NumFuncs = 2; NumInp = 1; NumOutp = 1; break;
		case 145:  NumFuncs = 2; ConfigIo = true;					break;
		case 2145: NumFuncs = 2; ConfigIo = true;					break;
		case 205:  NumFuncs = 1; NumInp = 0; NumOutp = 1; break;
		case 160:  NumFuncs = 3; ConfigIo = true;					break;	// Temporary IP port not used
		case 225:  NumFuncs = 1; NumInp = 1; NumOutp = 0; break;
		case 245:  NumFuncs = 2; NumInp = 1; NumOutp = 1; break;
    case 545:  NumFuncs = 2; NumInp = 1; NumOutp = 1; break;
		}
		// Configurable IO?
		if (ConfigIo)
		{
			// Count number of inputs and outputs
			for (Int ChanIdx=0; ChanIdx< NumFuncs; ChanIdx++) {
				Int IoConfig = 0;
				__DTAPI::GetIoConfigViaDta1xx(fdDriver, ChanIdx, IoConfig);
				if (IoConfig == DTAPI_IOCONFIG_INPUT)
					NumInp++;
				else if (IoConfig == DTAPI_IOCONFIG_OUTPUT)
					NumOutp++;
			}
			// Recalculate NumFuncs i.b.o. DTA-105
			NumFuncs = NumInp + NumOutp;
		}
								
		DvcDesc.m_NumHwFuncs = NumFuncs;
		DvcDesc.m_NumDtInpChan = NumInp;
		DvcDesc.m_NumDtOutpChan = NumOutp; 

		// Get serial number from EEPROM
		// Use a temporary device structure
		DtDeviceInt  DtDvcTemp;
		DtDvcTemp.m_Attached = true;
		DtDvcTemp.m_DvcDesc = DvcDesc;
		// Duplicate handle from DtDevice object to TsInpChannel object.
		int fd = dup(fdDriver);
		DtDvcTemp.m_fdDriver = fd;
		DvcDesc.m_Serial = DtDvcTemp.GetSerialFromVpd();

		// Perform action dependent on kind of scan
		switch (ScanType) {

		case SCAN_FULL:
			// Copy device data to DtDeviceDesc structure, unless buffer is too small.
			if (NumEntries >= NumEntriesResult) {
				*pDvcDesc = DvcDesc;	// Copy device descriptor to output array
				pDvcDesc++;				// Go to next DtDeviceDesc
			}
			break;

		case ATTACH_TO_SERIAL:
			// Check serial number
			if (SerialNumber == DvcDesc.m_Serial)
				DeviceFound = true;
			break;

		case ATTACH_TO_SLOT:
			// Check whether this is the right PCI bus and slot number
			if (DvcDesc.m_PciBusNumber==PciBusNumber &&	DvcDesc.m_SlotNumber==SlotNumber)
				DeviceFound = true;
			break;

		case ATTACH_TO_TYPE:
			// If this is the right type and relative device number, we're done
			if (TypeNumber == DvcDesc.m_TypeNumber)
				if (NumFound++ == DeviceNo)
					DeviceFound = true;
			break;
		}
		// If we are attaching and the right device has been found, we can break
		// the scanning loop (device-driver handle remains open).
		if (ScanType!=SCAN_FULL && DeviceFound)
			break;

		// Avoid resource leak
		close(fdDriver);
	}

	//.-.-.-.-.-.-.-.-.-.-.-.-.- Post-Processing for Full Scan -.-.-.-.-.-.-.-.-.-.-.-.-.-

	if (ScanType == SCAN_FULL) {
		// Full scan: did we have enough buffer space?
		if (NumEntriesResult > NumEntries)
			return DTAPI_E_BUF_TOO_SMALL;		// No...
		return DTAPI_OK;
	}

	//.-.-.-.-.-.-.-.-.-.-.-.-.-.- Post-Processing for Attach -.-.-.-.-.-.-.-.-.-.-.-.-.-.

	// Error checks
	if (NumEntriesResult == 0) {
		if (ScanType == ATTACH_TO_SLOT)
			return DTAPI_E_NO_DTA_CARD;			// No DEKTEC PCI cards at all
		else
			return DTAPI_E_NO_DEVICE;			// No DEKTEC devices at all
	}
	if (!DeviceFound)
		return DTAPI_E_NO_SUCH_DEVICE;

	// Check whether driver version is sufficient for use with this version of DtAPI
	int  vMajor, vMinor, vBugFix, vBuild;
	// NOTE: call to ::DtapiGetDeviceDriverVersion is not efficient, we already
	// have an open handle
	::DtapiGetDeviceDriverVersion(DvcDesc.m_Category, vMajor, vMinor, vBugFix, vBuild);
	dr = ::DtapiCheckDeviceDriverVersion(DvcDesc.m_Category, vMajor, vMinor, vBugFix, vBuild);
	if (dr != DTAPI_OK) {
		// Error => Clean up
		close(fdDriver);
		return dr;
	}

	// All is fine!
	// Initialise device structure
	pDvc->m_Attached = true;
	pDvc->m_DvcDesc = DvcDesc;
	pDvc->m_fdDriver = fdDriver;

	// Device-category-dependent processing
	if (pDvc->Category() == DTAPI_CAT_PCI) {
		// Nothing to be done
	}
	else if (pDvc->Category() == DTAPI_CAT_USB) {
		// Nothing to be done
	}
	else
		return DTAPI_E_INTERNAL;		// Error in device category

	return DTAPI_OK;
}

//=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ BACKWARD COMPATIBILITY +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=


//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtapiPciScan -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Scan hardware functions available on the DTA-xxx cards in the current system.
// DtapiPciScan is included for backwards compatiblity and will not list DTU-2XX
// devices. Replaced by DtapiDeviceScan.
//
DTAPI_RESULT  DtapiPciScan(
	int  NumFuncEntries,			// Number of function entries in pHwFuncs
	int&  NumFuncEntriesResult,		// #Functions found or #entries required in pHwFuncs
	DtapiHwFunc*  pHwFuncs)			// Hardware-function descriptions
{
	// Let's be cautious
	if (pHwFuncs==NULL) { return DTAPI_E_BUF_TOO_SMALL; }

	//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Device Scan -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

	char  DevicePath[32];
	DtapiHwFunc*  pHwf = pHwFuncs;
	NumFuncEntriesResult = 0;

	for (int nDeviceIndex=0; nDeviceIndex<10; nDeviceIndex++) {

		// Construct path and try to open device
		sprintf(DevicePath, "%s%d", c_Dta1xxPrefix, nDeviceIndex);
	    int  fdDta1xx = open(DevicePath, O_RDONLY | O_WRONLY);
		if ( fdDta1xx < 0 ) {
			continue;
		}

		// Found a new DTA-1xx PCI card

		// Get location info
		DTA1XX_IOCTL_DATA iod;
		memset(&iod, 0, sizeof(DTA1XX_IOCTL_DATA));
		int nResult = ioctl(fdDta1xx, DTA1XX_IOCTL_GET_LOCATION_INFO,  &iod);

		// If DeviceIoControl fails => do not list this hardware function
		if (nResult!=0 ) {
			close(fdDta1xx);		// Do not leak handles
			continue;
		}

		// Determine number of hardware functions on this board
		int  nTypeNumber = 	DtapiDeviceId2TypeNumber(iod.m_Data.m_LocInfo.m_nSubSystemId);
		int  nNumHwFuncs = 0;
		switch (nTypeNumber) {
		case 100: nNumHwFuncs = 1; break;
		case 102: nNumHwFuncs = 1; break;
		case 105: nNumHwFuncs = 2; break;
		case 120: nNumHwFuncs = 1; break;
		case 122: nNumHwFuncs = 1; break;
		case 140: nNumHwFuncs = 2; break;
		}

		// Loop through hardware functions
		for (int nIdx=0; nIdx<nNumHwFuncs; nIdx++) {

			// Copy to DtapiHwFunc structure, unless buffer is too small.
			NumFuncEntriesResult++;
			if (NumFuncEntries >= NumFuncEntriesResult) {
				pHwf->m_nPciBusNumber	= iod.m_Data.m_LocInfo.m_nBusNumber;
				pHwf->m_nSlotNumber	= iod.m_Data.m_LocInfo.m_nSlotNumber;
				pHwf->m_nDeviceId	= iod.m_Data.m_LocInfo.m_nDeviceId;
				pHwf->m_nVendorId	= iod.m_Data.m_LocInfo.m_nVendorId;
				pHwf->m_nSubsystemId	= iod.m_Data.m_LocInfo.m_nSubSystemId;
				pHwf->m_nSubVendorId	= iod.m_Data.m_LocInfo.m_nSubVendorId;

				// Type number is derived from Device ID
				pHwf->m_nTypeNumber		= nTypeNumber;

				// Decode hardware-function type
				pHwf->m_nHwFuncType	= 0;
				if (nTypeNumber==100 || nTypeNumber==102 || nTypeNumber==105 || (nTypeNumber==140 && nIdx==0))
					pHwf->m_nHwFuncType	= DTAPI_TS_OUTPUT;

				if (nTypeNumber==120 || nTypeNumber==122 || (nTypeNumber==140 && nIdx==1))
					pHwf->m_nHwFuncType	= DTAPI_TS_INPUT;

				pHwf->m_nHwFuncFlags	= 0;

				if ( nTypeNumber==105 ) {
					pHwf->m_nIndexOnCard = nIdx;
				} else {
					pHwf->m_nIndexOnCard = 0;
				}

				pHwf++;
			}
		}

		close(fdDta1xx);
	}

	// Did we have enough buffer space?
	if (NumFuncEntriesResult > NumFuncEntries) {
		return DTAPI_E_BUF_TOO_SMALL;		// No...
	}

	return DTAPI_OK;
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtapiResult2Str -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Convert DTAPI_RESULT value to a string
//
const char*  DtapiResult2Str(
        DTAPI_RESULT  DtapiResult)
{
        switch (DtapiResult) {
        case DTAPI_OK:					return "DTAPI_OK"; break;
        case DTAPI_E_ATTACHED:			return "DTAPI_E_ATTACHED"; break;
        case DTAPI_E_BUF_TOO_SMALL:		return "DTAPI_E_BUF_TOO_SMALL"; break;
        case DTAPI_E_DEV_DRIVER:		return "DTAPI_E_DEV_DRIVER"; break;
        case DTAPI_E_DEVICE:			return "DTAPI_E_DEVICE"; break;
        case DTAPI_E_DST_MAC_ADDR:		return "DTAPI_E_DST_MAC_ADDR"; break;
        case DTAPI_E_DRIVER_INCOMP:		return "DTAPI_E_DRIVER_INCOMP"; break;
        case DTAPI_E_EEPROM_FULL:		return "DTAPI_E_EEPROM_FULL"; break;
        case DTAPI_E_EEPROM_READ:		return "DTAPI_E_EEPROM_READ"; break;
        case DTAPI_E_EEPROM_WRITE:		return "DTAPI_E_EEPROM_WRITE"; break;
        case DTAPI_E_EEPROM_FORMAT:		return "DTAPI_E_EEPROM_FORMAT"; break;
        case DTAPI_E_EMBEDDED:			return "DTAPI_E_EMBEDDED"; break;
        case DTAPI_E_FIFO_FULL:			return "DTAPI_E_FIFO_FULL"; break;
        case DTAPI_E_IDLE:				return "DTAPI_E_IDLE"; break;
        case DTAPI_E_IN_USE:			return "DTAPI_E_IN_USE"; break;
        case DTAPI_E_INTERNAL:			return "DTAPI_E_INTERNAL"; break;
        case DTAPI_E_INVALID_BANDWIDTH:	return "DTAPI_E_INVALID_BANDWIDTH"; break;
        case DTAPI_E_INVALID_BUF:		return "DTAPI_E_INVALID_BUF"; break;
        case DTAPI_E_INVALID_CGMODE:	return "DTAPI_E_INVALID_CGMODE"; break;
        case DTAPI_E_INVALID_CONSTEL:	return "DTAPI_E_INVALID_CONSTEL"; break;
        case DTAPI_E_INVALID_FLAGS:		return "DTAPI_E_INVALID_FLAGS"; break;
        case DTAPI_E_INVALID_GUARD:		return "DTAPI_E_INVALID_GUARD"; break;
        case DTAPI_E_INVALID_HEADER:	return "DTAPI_E_INVALID_HEADER"; break;
        case DTAPI_E_INVALID_INTERLVNG:	return "DTAPI_E_INVALID_INTERLVNG"; break;
        case DTAPI_E_INVALID_MODE:		return "DTAPI_E_INVALID_MODE"; break;
        case DTAPI_E_INVALID_RATE:		return "DTAPI_E_INVALID_RATE"; break;
        case DTAPI_E_INVALID_ROLLOFF:	return "DTAPI_E_INVALID_ROLLOFF"; break;
        case DTAPI_E_INVALID_SIZE:		return "DTAPI_E_INVALID_SIZE"; break;
        case DTAPI_E_INVALID_TRANSMODE:	return "DTAPI_E_INVALID_TRANSMODE"; break;
        case DTAPI_E_INVALID_TSTYPE:	return "DTAPI_E_INVALID_TSTYPE"; break;
        case DTAPI_E_INSUF_LOAD:		return "DTAPI_E_INSUF_LOAD"; break;
        case DTAPI_E_KEYWORD:			return "DTAPI_E_KEYWORD"; break;
        case DTAPI_E_LOCKED:			return "DTAPI_E_LOCKED"; break;
        case DTAPI_E_MULTICASTJOIN:		return "DTAPI_E_MULTICASTJOIN"; break;
        case DTAPI_E_NDIS_DRIVER:		return "DTAPI_E_NDIS_DRIVER"; break;
        case DTAPI_E_NO_DEVICE:			return "DTAPI_E_NO_DEVICE"; break;
        case DTAPI_E_NO_DT_INPUT:		return "DTAPI_E_NO_DT_INPUT"; break;
        case DTAPI_E_NO_DT_OUTPUT:		return "DTAPI_E_NO_DT_OUTPUT"; break;
        case DTAPI_E_NO_IPPARS:			return "DTAPI_E_NO_IPPARS"; break;
        case DTAPI_E_NO_LINK:			return "DTAPI_E_NO_LINK"; break;
        case DTAPI_E_NO_LOOPBACK:		return "DTAPI_E_NO_LOOPBACK"; break;
        case DTAPI_E_NO_SUCH_DEVICE:	return "DTAPI_E_NO_SUCH_DEVICE"; break;
        case DTAPI_E_NO_SUCH_INPUT:		return "DTAPI_E_NO_SUCH_INPUT"; break;
        case DTAPI_E_NO_SUCH_OUTPUT:	return "DTAPI_E_NO_SUCH_OUTPUT"; break;
        case DTAPI_E_NO_SUCH_PORT:		return "DTAPI_E_NO_SUCH_PORT"; break;
        case DTAPI_E_NO_TSRATE:			return "DTAPI_E_NO_TSRATE"; break;
        case DTAPI_E_NO_VALID_CALDATA:	return "DTAPI_E_NO_VALID_CALDATA"; break;
        case DTAPI_E_NOT_ATTACHED:		return "DTAPI_E_NOT_ATTACHED"; break;
        case DTAPI_E_NOT_FOUND:			return "DTAPI_E_NOT_FOUND"; break;
        case DTAPI_E_NOT_IDLE:			return "DTAPI_E_NOT_IDLE"; break;
        case DTAPI_E_NOT_SUPPORTED:		return "DTAPI_E_NOT_SUPPORTED"; break;
        case DTAPI_E_OUT_OF_MEM:		return "DTAPI_E_OUT_OF_MEM"; break;
        case DTAPI_E_TOO_LONG:			return "DTAPI_E_TOO_LONG"; break;
        case DTAPI_E_UNDERFLOW:			return "DTAPI_E_UNDERFLOW"; break;
        case DTAPI_E_WINSOCK:			return "DTAPI_E_WINSOCK"; break;
        }
        return "??? UNKNOWN DTAPI_RESULT ???";
}
