//*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* Utility.cpp *#*#*#*#*#*#*#*#* (C) 2000-2004 DEKTEC
//
// DTAPI
//
// Implementation of global utility functions.

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- License -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// This program is free software; You can freely use/modify this source code in any way
// you desire, including for commercial applications.
//

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Change History -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//	TD	2007.01.31	Added support for SetIoConfigViaDta1xx and GetIoConfigViaDta1xx
//	JH	2005.07.13	Added support for cards without subsystem (non-PLX)
//	MG	19-11-2002	Created	based on DTAPI for windows

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Include files -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
#include "StdAfx.h"
#include <stdlib.h>

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Constants -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtapiDeviceId2TypeNumber -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Returns the typenumber of a DTA1xx card based on its subsystem ID or device ID.
// (formerly DtapiSubsystemId2TypeNumber())
//
int DtapiDeviceId2TypeNumber(
	int  nDeviceId)	// Subsystem ID or DeviceId of the DTA-1xx card
{
	// For the moment, this is simple
	return nDeviceId - (nDeviceId < 0xD100 ? 0 : (0xD100-100));
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- TxMode2PacketSize -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Computes the packet size based on the Transmit mode
//
int TxMode2PacketSize(
		int  TxPacketMode)		// Packet transmit mode (TxMode, BurstMode)
{
	switch (TxPacketMode & DTAPI_TXMODE_MASK) {
	case DTAPI_TXMODE_188:		return 188;
	case DTAPI_TXMODE_192:		return 192;
	case DTAPI_TXMODE_204:		return 204;
	case DTAPI_TXMODE_ADD16:	return 204;
	case DTAPI_TXMODE_MIN16:	return 204;
	case DTAPI_TXMODE_RAW:		return 188;
	default:					return 188;		// Should not occur
	}
}



//-.-.-.-.-.-.-.-.-.-.-.-.-.-.- __DTAPI::ResetOutputChannel -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Device IO-Control to issue hardware reset of output channel.
// This call will also reset the status flags.
//
DTAPI_RESULT  __DTAPI::ResetOutputChannel(
        HANDLE  hDriver,	// [in] Handle to device driver
        Int  PortIndex,		// [in] Zero-based port index
        Int  ResetMode)		// [in] DTAPI_FIFO_RESET / DTAPI_FULL_RESET
{
      DTA1XX_IOCTL_DATA  iod;
      memset(&iod, 0, sizeof(DTA1XX_IOCTL_DATA));
      iod.m_nChIdx = PortIndex;
      iod.m_Data.m_uiValue = ResetMode;

      DTAPI_RESULT dr = ioctl(hDriver, DTA1XX_TX_IOCTL_RESET, &iod);
      // IO control failed??
      if (dr!=DTAPI_OK)
          return DTAPI_E_DEV_DRIVER;
      return DTAPI_OK;
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.- __DTAPI::GetIoConfigViaDta1xx -.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Internal DTAPI function for reading the I/O configuration of a physical port in 
// device scan.
//
// RETURN VALUE:	true	I/O configuration read successfully
//					false	Error in reading I/O configuration
//
Bool  __DTAPI::GetIoConfigViaDta1xx(
	HANDLE  hDriver,	// [in] Handle to device driver
	Int  PortIndex,		// [in] Zero-based port index
	Int&  IoConfig)		// [out] I/O configuration: DTAPI_IOCONFIG_XXX constants
{
	IoConfig = 0;
	
	DTA1XX_IOCTL_DATA  iod;
	memset(&iod, 0, sizeof(DTA1XX_IOCTL_DATA));
	iod.m_nChIdx = PortIndex;

	DTAPI_RESULT dr = ioctl(hDriver, DTA1XX_IOCTL_GET_IOCONFIG,  &iod);

      if (dr!=DTAPI_OK)
		return false;

	// Copy back IoConfig
	IoConfig = iod.m_Data.m_uiValue;
	return true;
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.- __DTAPI::SetIoConfigViaDta1xx -.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Internal DTAPI function for writing the I/O configuration of a physical port in 
// device scan.
//
// RETURN VALUE:	true	I/O configuration written successfully
//					false	Error in writing I/O configuration
//
Bool  __DTAPI::SetIoConfigViaDta1xx(
		HANDLE  hDriver,	// [in] Handle to device driver
		Int  PortIndex,		// [in] Zero-based port index
		Int  IoConfig)		// [in] I/O configuration: DTAPI_IOCONFIG_XXX constants
{
	DTA1XX_IOCTL_DATA  iod;
	memset(&iod, 0, sizeof(DTA1XX_IOCTL_DATA));
	iod.m_nChIdx = PortIndex;
	iod.m_Data.m_uiValue = IoConfig;

	DTAPI_RESULT dr =  ioctl(hDriver, DTA1XX_IOCTL_SET_IOCONFIG,  &iod);
	if (dr!=DTAPI_OK)
		return false;

	return true;
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- __DTAPI::SetTxControl -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Internal DTAPI function for invoking IOCTL_DTA1XX_SET_TX_CONTROL
//
DTAPI_RESULT  __DTAPI::SetTxControl(
        HANDLE  hDriver,	// [in] Handle to device driver
        Int  Category,		// [in] Device category
        Int  PortIndex,		// [in] Zero-based port index
        Int  TxControl)		// [in] Transmit-control state
{
    DTAPI_RESULT dr;

    // Device-category dependent processing
    if (Category == DTAPI_CAT_PCI)
    {
      DTA1XX_IOCTL_DATA  iod;
      memset(&iod, 0, sizeof(DTA1XX_IOCTL_DATA));
      iod.m_nChIdx = PortIndex;
      iod.m_Data.m_uiValue = TxControl;

      dr = ioctl(hDriver, DTA1XX_TX_IOCTL_SET_TX_CONTROL, &iod);
      // IO control failed??
      if (dr!=DTAPI_OK)
        return dr;
    }
    else if (Category == DTAPI_CAT_USB) {

        DTU2XX_IOCTL_DATA  iod;
        memset(&iod, 0, sizeof(DTU2XX_IOCTL_DATA));
        iod.m_nChIdx = PortIndex;
        iod.m_Data.m_uiValue = TxControl;

        dr = ioctl(hDriver, DTU2XX_TX_IOCTL_SET_TX_CONTROL, &iod);
        // IO control failed??
        if (dr!=DTAPI_OK) {
                return dr;
        }
    }
    else
         return DTAPI_E_INTERNAL;			// Unknown device category

    return DTAPI_OK;
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- __DTAPI::ReadRegister -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
DTAPI_RESULT __DTAPI::ReadRegister(
    HANDLE hDriver,
    Int PortIndex,
    Int Address,
    Int& RegValue)
{
    DTAPI_RESULT dr;
    DTA1XX_IOCTL_DATA  iod;
    memset(&iod, 0, sizeof(DTA1XX_IOCTL_DATA));
    iod.m_nChIdx = 0;
	iod.m_Data.m_RegCtrl.m_Address = Address;

	dr = ioctl(hDriver, DTA1XX_IOCTL_READ_REG, &iod);
    // IO control failed??
    if (dr!=DTAPI_OK)
		return dr;

	RegValue = iod.m_Data.m_RegCtrl.m_RegValue;

	return DTAPI_OK;
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- __DTAPI::WriteRegister -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
DTAPI_RESULT __DTAPI::WriteRegister(
    HANDLE hDriver,
    Int PortIndex,
    Int Address,
    Int RegValue)
{
  DTAPI_RESULT dr;

	DTA1XX_IOCTL_DATA  iod;
    memset(&iod, 0, sizeof(DTA1XX_IOCTL_DATA));
    iod.m_nChIdx = PortIndex;
	iod.m_Data.m_RegCtrl.m_Address = Address;
	iod.m_Data.m_RegCtrl.m_RegValue = RegValue;

	dr = ioctl(hDriver, DTA1XX_IOCTL_WRITE_REG, &iod);
    // IO control failed??
    if (dr!=DTAPI_OK)
		return dr;

	return DTAPI_OK;
}

int min(int a, int b)
{
  if (a<b)
    return a;
  else
    return b;
}

void Sleep(int ms)
{
  usleep(ms*1000);
}
