//#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* Vpd.c *#*#*#*#*#*#*#*#*#*# (C) 2000-2005 DEKTEC
//
// Vital Product Data (VPD) read/write routines.

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- License -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// This program is free software; You can freely use/modify this source code in any way
// you desire, including for commercial applications.
//

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Change History -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//	KvR	14-08-2006	DTA-105 registers changed, Vpd-data handling changed
//	JH	16-08-2004	Removed some unused variables
//	MG	2003.04.23	Created. Based on windows driver code

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Include files -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
#include "../Include/Dta1xx.h"

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Tunable Constants -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
#define VPD_WRARR_MAX_RETRIES	5		// Max. #times an array write is retried
#define VPD_WRITE_MAX_RETRIES	5		// Max. #times a VPD write is retried


//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Local Forwards -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
int  Dta1xxVpdFindResource(DTA1XX_FDO*, char*, unsigned char*, unsigned int, unsigned char**);
int  Dta1xxVpdIsValidResource(unsigned char*, unsigned char*, unsigned int);
int  Dta1xxVpdWriteArray(DTA1XX_FDO*, unsigned int*, unsigned int, unsigned int);
int  Dta1xxVpdWriteWordWithCheck(DTA1XX_FDO*, unsigned int, unsigned int);


//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Dta1xxVpdReadItem -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Read VPD item into a caller-allocated string buffer.
// This routine is used to read PCI-card specific configuration data at
// driver-initialisation time .
//
// PRE: EEPROM contents have been read into VPD-EEPROM cache (m_pVpdEeprom)
//
int  Dta1xxVpdReadItem(
	DTA1XX_FDO*  pFdo,			// Device extension
	char*  pKeyword,				// Name of VPD item to be read
	char*  pVpdItem)				// String buffer for storing VPD item
{
	unsigned int  i, Length, RoLength;
	char  *pFrom, *pTo;					// Helper character pointers
	unsigned char*  pVpd;				// Byte pointer in VPD data
	unsigned char*  pVpdRoEnd;			// Pointer one after RO section
	unsigned char*  pResource;			// Resource to be read

	// Add parameter checks here

	// Compute a number of helper variables
	pVpdRoEnd = (unsigned char*)&pFdo->m_pVpdEeprom[PLX_VPD_ID_START + PLX_VPD_ID_RO_LENGTH];

	// Initialisation
	pTo = pVpdItem;				// For later copying to VPD item

	// Start at beginning of VPD-ID + read-only section
	pVpd = (unsigned char*)&pFdo->m_pVpdEeprom[PLX_VPD_ID_START];
	if (pVpd[0] != 0x82) {
		DTA1XX_LOG1( KERN_DEBUG, "Dta1xxVpdReadItem: ERROR: VPD-ID section"
				"starts with 0x%02x i.s.o. 0x82", pVpd[0] );
		return -EFAULT;
	}

	// NO SUPPORT FOR READING VPD-ID YET

	// Move to start of VPD read-only section
	pVpd += 3 + pVpd[1] + (pVpd[2]<<8);

	// Consistency checks
	if (pVpd >= pVpdRoEnd) {
		DTA1XX_LOG( KERN_DEBUG, "Dta1xxVpdReadItem: ERROR: VPD-RO section starts"
				"beyond max. end address" );
		return -EFAULT;
	}
	if (pVpd[0] != 0x90) {
		DTA1XX_LOG1( KERN_DEBUG, "Dta1xx: Dta1xxVpdReadItem: ERROR: VPD-RO section"
			  "starts with 0x%02x i.s.o. 0x90", pVpd[0] );
		return -EFAULT;
	}

	// Length of RO section
	RoLength = (unsigned int)(pVpdRoEnd - pVpd);

	// Check whether we can find resource in RO section
	if (Dta1xxVpdFindResource(pFdo, pKeyword, pVpd, RoLength, &pResource)) {
 		pFrom = pResource + 3;
		Length = pResource[2];
		if (Length > VPD_ITEM_MAX_LENGTH) {
			Length = VPD_ITEM_MAX_LENGTH;	// Protect against invalid content of EEPROM
		}
		for (i=0; i<Length; i++) {
			*pTo++ = *pFrom++;
		}
		*pTo = '\0';
#ifdef DEBUG
		DTA1XX_LOG2( KERN_DEBUG, "Dta1xxVpdReadItem: Found VPD item '%s' in RO"
			   "section, value='%s'", pKeyword, pVpdItem );
#endif
		return 0;
	}

	// Check whether we can find resource in RW section
	pVpd = (unsigned char*)&pFdo->m_pVpdEeprom[PLX_VPD_RW_START];
	if (Dta1xxVpdFindResource(pFdo, pKeyword, pVpd, PLX_VPD_RW_LENGTH, &pResource)) {
 		pFrom = pResource + 3;
		Length = pResource[2];
		if (Length > VPD_ITEM_MAX_LENGTH) {
			Length = VPD_ITEM_MAX_LENGTH;	// Protect against invalid content of EEPROM
		}
		for (i=0; i<Length; i++) {
			*pTo++ = *pFrom++;
		}
		*pTo = '\0';
#ifdef DEBUG
		DTA1XX_LOG2( KERN_DEBUG, "Dta1xxVpdReadItem: Found VPD item '%s' in RW section, "
				 "value='%s'", pKeyword, pVpdItem );
#endif
		return 0;
	}

	// Not found
	DTA1XX_LOG1( KERN_DEBUG, "Dta1xxVpdReadItem: Can't find VPD item '%s'", pKeyword );

	return -EFAULT;
}


//=+=+=+=+=+=+=+=+=+=+=+=+=+=+ Reading and Writing Raw Bytes +=+=+=+=+=+=+=+=+=+=+=+=+=+=+


//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Dta1xxVpdRead -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Read an array of bytes from the serial EEPROM.
//
int  Dta1xxVpdRead(
	DTA1XX_FDO*  pFdo,			// Functional device object, representing the DTA-1xx card
	unsigned char*	pBuf,		// Buffer for data read from serial EEPROM
	unsigned int  StartAddr,	// Start address (BYTE address)
	unsigned int  Length)		// Number of BYTES to read
{
	unsigned int  RelAddr, WordAddr, WordAddrLast;
	unsigned int  Word16, VpdWord;
	int  RetValue;

#ifdef DEBUG
	DTA1XX_LOG2( KERN_DEBUG, "Dta1xxVpdRead StartAddr=0x%03x Length=%d", StartAddr, Length );
#endif

//        if (pFdo->m_nTypeNumber == 105 || pFdo->m_nTypeNumber == 545)
//            return 0;

	// Check parameters
	if (StartAddr >= DTA1XX_VPD_BUF_SIZE) {
		DTA1XX_LOG1( KERN_DEBUG, "Dta1xxVpdRead: StartAddr=0x%03x INVALID", StartAddr );
		return -EFAULT;
	}
	if (StartAddr + Length > DTA1XX_VPD_BUF_SIZE) {
		DTA1XX_LOG1( KERN_DEBUG, "Dta1xxVpdRead: Length=%d INVALID", Length );
		return -EFAULT;
	}

	// Let's be robust
	if (Length == 0) {
		return 0;
	}

	if (!pFdo->m_nUsesPlxChip) {
		// First read complete EEPROM
		// then use Dta1xxVpdReadFromCache to read from cache
		if (pFdo->m_UseFpgaForVpd) {
			// Read from EEPROM
				Dta1xxEepromReadLastPage(pFdo);
			return Dta1xxVpdReadFromCache(pFdo, pBuf, StartAddr, Length);
		} else {
			// Read from EEPROM
			Dta1xxEepromSpiRead(pFdo);
			return Dta1xxVpdReadFromCache(pFdo, pBuf, StartAddr, Length);
		}
		DTA1XX_LOG( KERN_DEBUG, "Dta1xxVpdRead: Not supported" );
		return -EFAULT;
	}

	// Compute start- and last 32-bit word addresses
	WordAddr = StartAddr / 4;
	RelAddr = StartAddr % 4;
	WordAddrLast = (StartAddr + Length - 1) / 4;

	// Read 32-bit word by word and extract bytes
	for (; WordAddr<=WordAddrLast; WordAddr++) {

		// Read two 16-bit words, into 32-bit VpdWord
		RetValue = Dta1xxEepromReadWord16(pFdo, WordAddr*2, &Word16);
		if (RetValue != 0) {
			return (RetValue);
		}
		VpdWord = Word16<<16;
		RetValue = Dta1xxEepromReadWord16(pFdo, WordAddr*2+1, &Word16);
		if (RetValue != 0) {
			return (RetValue);
		}
		VpdWord |= Word16;

		// A little bit of magic to convert 32-bit word to bytes
		while (RelAddr<4 && Length!=0) {
			*pBuf++ = (unsigned char) ((VpdWord >> (8 * (3-RelAddr))) & 0xFF);
			RelAddr++;
			Length--;
		}
		RelAddr = 0;
	}
	return 0;
}


//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Dta1xxVpdReadFromCache -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Read an array of bytes from the EEPROM cache.
//
int  Dta1xxVpdReadFromCache(
	DTA1XX_FDO*  pFdo,			// Device extension
	unsigned char*	pBuf,					// Buffer for data read from serial EEPROM
	unsigned int  StartAddr,				// Start address (BYTE address)
	unsigned int  Length)					// Number of BYTES to read
{
	unsigned int  ByteIndex;
	unsigned int Addr;

#ifdef DEBUG
	DTA1XX_LOG2( KERN_DEBUG, "Dta1xxVpdRead: StartAddr=0x%03x Length=%d", StartAddr, Length );
#endif

	// Check parameters
	if (StartAddr >= DTA1XX_VPD_BUF_SIZE) {
		DTA1XX_LOG1( KERN_DEBUG, "Dta1xxVpdRead: StartAddr=0x%03x INVALID", StartAddr );
		return -EFAULT;
	}
	if (StartAddr + Length > DTA1XX_VPD_BUF_SIZE) {
		DTA1XX_LOG1( KERN_DEBUG, "Dta1xxVpdRead: Length=%d INVALID", Length );
		return -EFAULT;
	}

	// Let's be robust
	if (Length == 0) {
		return 0;
	}

	// Read from VPD cache for all cards
	Addr = StartAddr;

	for (ByteIndex=0; ByteIndex<Length; ByteIndex++) {
		*pBuf++ = pFdo->m_pVpdEeprom[Addr++];
	}
	return 0;
}


//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Dta1xxVpdWrite -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Write an array of bytes to the serial EEPROM.
// Writing is complicated by the fact that we can only write 32-bit words.
//
int  Dta1xxVpdWrite(
	DTA1XX_FDO*  pFdo,			// Functional device object, representing the DTA-1xx card
	unsigned char*  pBuf,		// Buffer with data to write to serial EEPROM
	unsigned int  StartAddr,	// Start address (BYTE address)
	unsigned int  Length)		// Number of BYTES to read
{
	int  Try;
	unsigned int  ByteIndex, Mask, RelAddr, RelAddrStart, RelAddrLast;
	unsigned int  WordAddrStart, WordAddrLast;
	unsigned int  WordBuf[DTA1XX_VPD_MAX_TRANSFER / 2];
	unsigned int*  pWord;
	int  RetValue;

#ifdef DEBUG
	DTA1XX_LOG3( KERN_DEBUG, "Dta1xxVpdWrite: StartAddr=0x%03x Length=%d pBuf=0x%p",
			 StartAddr, Length, pBuf );
#endif

	// Check parameters
	if (StartAddr >= DTA1XX_VPD_BUF_SIZE) {
		DTA1XX_LOG1( KERN_DEBUG, "Dta1xx: Dta1xxVpdWrite: StartAddr=0x%03x INVALID", StartAddr );
		return -EFAULT;
	}
	if (StartAddr + Length > DTA1XX_VPD_BUF_SIZE) {
		DTA1XX_LOG1( KERN_DEBUG, "Dta1xx: Dta1xxVpdWrite: Length=%d INVALID", Length );
		return -EFAULT;
	}

	// Let's be robust
	if (Length == 0) {
		return 0;
	}

	if( pFdo->m_nUsesPlxChip ) {

		//-.-.-.-.-.-.-.-.-.- Convert byte array to array of 16-bit words -.-.-.-.-.-.-.-.-.-.

		// Compute start and end 16-bit word addresses
		WordAddrStart = StartAddr / 2;
		RelAddrStart = StartAddr % 2;
		WordAddrLast = (StartAddr + Length - 1) / 2;
		RelAddrLast = (StartAddr + Length - 1) % 2;

		// Read first 16-bit word if we have a "partial" word at the beginning
		if (RelAddrStart != 0) {
			RetValue = Dta1xxEepromReadWord16(pFdo, WordAddrStart, &WordBuf[WordAddrStart]);
			if (RetValue != 0) {
				return RetValue;
			}
		}

		// Read last word if we have a "partial" word at the end
		if (RelAddrLast != 3) {
			RetValue = Dta1xxEepromReadWord16(pFdo, WordAddrLast, &WordBuf[WordAddrLast]);
			if (RetValue != 0) {
				return RetValue;
			}
		}

		// Construct array of 16-bit words to be written to the serial EEPROM
		pWord = &WordBuf[WordAddrStart];
		RelAddr = RelAddrStart;
		for (ByteIndex=0; ByteIndex<Length; ByteIndex++) {
			Mask = 0xFF << (8 * (1-RelAddr));
			*pWord &= ~Mask;
			*pWord |= pBuf[ByteIndex] << (8 * (1-RelAddr));
			*pWord &= 0xFFFF;
			if (++RelAddr == 2) {
				RelAddr = 0;
				pWord++;
			}
		}

		//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Program serial EEPROM -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

		// Change "write-protected area" address in PCI 9054
		//	WRITE_UINT16(p905XRegs, PCI905X_PROT_AREA, WordAddrStart);

		// Set PE (Serial EEPROM Program Enable)
		//	pFdo->m_pDta1xxGen->m_GenControl.m_PE = 1;

		// Write until success, with a maximum number of retries

		// TODO: Read from cache, write to EEPROM, read back and store in cache
		for (Try=0; Try<VPD_WRARR_MAX_RETRIES; Try++) {
			RetValue = Dta1xxVpdWriteArray(pFdo, WordBuf, WordAddrStart, WordAddrLast);
			if (RetValue==0) {
				break;
			}
		}

		// Clear PE (Serial EEPROM Program Enable)
		// pFdo->m_pDta1xxGen->m_GenControl.m_PE = 0;

		// Disable "write-protected area" address in PCI 9054
		// WRITE_UINT16(p905XRegs, PCI905X_PROT_AREA, 0x7F);

		// Read back written content into cache
		// TODO: Combine write and read back
		Dta1xxVpdRead(pFdo, (unsigned char*)(pFdo->m_pEepromCache)+StartAddr, StartAddr, Length);

		return RetValue;
	}

	if (pFdo->m_UseFpgaForVpd) {
		// Reprogram last sector with the VPD data
		RetValue = Dta1xxEepromWriteLastPage(pFdo);

		// Read back written content into cache
		Dta1xxEepromReadLastPage(pFdo);
		return RetValue;
	}

	RetValue = Dta1xxEepromSpiWrite(pFdo);

	// Read back written content into cache
	Dta1xxEepromSpiRead(pFdo);
	return RetValue;
}


//+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ Helpers +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=


//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Dta1xxVpdFindResource -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Keyword-based VPD-resource search in given VPD section (RO or RW)
//
int  Dta1xxVpdFindResource(
	DTA1XX_FDO*  pFdo,				// Device extension
	char*  pKeyword,					// Keyword to be found
	unsigned char*  pSection,					// VPD section (starts with VPD tag)
	unsigned int  SectLength,					// Real section length
	unsigned char**  ppResource)				// Points to VPD resource if keyword is found
{
	unsigned char*  pRes;			// Running resource pointer
	unsigned int  ResLength;		// Resource length
	
	pRes = pSection + 3;

	// Loop through resources
	while (Dta1xxVpdIsValidResource(pRes, pSection, SectLength)) {
//		KdPrint(("Dta1xx: Dta1xxVpdFindResource, checking '%c%c'\n", pRes[0], pRes[1]));
		if (strncmp((char*) pRes, (char*) pKeyword, 2) == 0) {
			*ppResource = pRes;
			return 1;
		}
		ResLength = pRes[2] + 3;
		pRes += ResLength;
	}

	// No more valid keywords => can't find keyword
	return 0;
}


//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Dta1xxVpdIsValidResource -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Check whether pResource points to a correctly formatted VPD resource.
//
int  Dta1xxVpdIsValidResource(
	unsigned char*  pResource,		// Resource to be checked for validity
	unsigned char*  pSection,		// Section in which keyword is located
	unsigned int    SectLength)		// Real section length (should correspond to pSection[1..2]
{
	if (pResource > pSection + SectLength - 3)
		return 0;
	if (!VALID_VPD_KEYW_CHAR(pResource[0]) || !VALID_VPD_KEYW_CHAR(pResource[1]))
		return 0;
	return (pResource + 3 + (unsigned int)(pResource[2])) <= (pSection + SectLength);
}


//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Dta1xxVpdWriteArray -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Write an array of 16-bit words to the serial EEPROM.
// Assumes PROT_AREA in PCI 9054 and PE in Altera have been enabled.
//
int  Dta1xxVpdWriteArray(
	DTA1XX_FDO*  pFdo,			// Functional device object, representing the DTA-1xx card
	unsigned int*  pWordBuf,	// 16-bit data to be written
	unsigned int  WordAddrStart,// First 16-bit word in pWordBuf to be written
	unsigned int  WordAddrLast)	// Last 16-bit word in pWordBuf to be written
{
	unsigned int  WordAddr, VpdWord;
	int  OkFlag = 1;
	int  RetValue;

	// Loop through 16-bit words.
	for (WordAddr=WordAddrStart; WordAddr<=WordAddrLast; WordAddr++) {

		// Read word first. Program only if word is required to change.
		RetValue = Dta1xxEepromReadWord16(pFdo, WordAddr, &VpdWord);
		if (RetValue != 0) {
			return RetValue;		// Reads must succeed
		}
		if (VpdWord == pWordBuf[WordAddr]) {
			DTA1XX_LOG2( KERN_DEBUG, "Address 0x%03x already contains 0x%08x",
						 WordAddr, VpdWord );
			continue;
		}

		DTA1XX_LOG2( KERN_DEBUG, "Program address 0x%03x with 0x%08x", WordAddr,
			         pWordBuf[WordAddr] );
		// Program 32-bit word to the serial EEPROM
		RetValue = Dta1xxVpdWriteWordWithCheck(pFdo, WordAddr, pWordBuf[WordAddr]);
		if ( RetValue != 0 ) {
			OkFlag = 0;
		}
	}
	return OkFlag ? 0 : -EFAULT;
}


//-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Dta1xxVpdWriteWordWithCheck -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Write one 16-bit word to the serial EEPROM and check whether the word has been written
// successfully.
// Retry max. VPD_WRITE_MAX_RETRIES times.
//
int  Dta1xxVpdWriteWordWithCheck(
	DTA1XX_FDO*  pFdo,			// Functional device object, representing the DTA-1xx card
	unsigned int  WordAddr,	// 16-bit word address
	unsigned int  DataWord)	// 16-bit data to be written
{
	int  Try;
	unsigned int  VpdWord;
	int  OkFlag = 0;
	int  RetValue;

	for (Try=0; Try<VPD_WRITE_MAX_RETRIES; Try++) {

		// Write word
		RetValue = 	Dta1xxEepromWriteWord16(pFdo, WordAddr, DataWord);
		if (RetValue != 0) {
			continue;					// Try again if write fails
		}

		// Read and check word
		RetValue = Dta1xxEepromReadWord16(pFdo, WordAddr, &VpdWord);
		if (RetValue != 0) {
			return RetValue;			// Reads must succeed
		}
		if (VpdWord == DataWord) {
			OkFlag = 1;
			break;
		}
		DTA1XX_LOG1( KERN_DEBUG, "Write fails: read-back value=0x%03x", VpdWord );
	}
	return OkFlag ? 0 : -EFAULT;
}
