//#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* Utility.c *#*#*#*#*#*#*#*#*# (C) 2000-2006 DekTec
//
// This code is the intellectual property of DekTec Digital Video B.V. and may not be
// distributed, in its original or in altered form, without the explicit permission of
// DekTec Digital Video B.V.

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Change History -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//  2006.10.01	MG	Clean-up of code
//  2006.08.14	KvR	Changed to new interface, added support for DTU-245, several bugfixes
//  2004.11.24  MG  Created

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Include files -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
#include "Dtu2xx.h"

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Dtu2xxTxMode2PacketSize -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Computes the packet size based on the packet transmit-mode
//
Int Dtu2xxTxMode2PacketSize(
    Int TxPacketMode)		// Packet transmit mode (TxMode, BurstMode)
{
    switch (TxPacketMode & DTU2XX_TXMODE_MASK)
    {
    case DTU2XX_TXMODE_188:		return 188;
    case DTU2XX_TXMODE_192:		return 192;
    case DTU2XX_TXMODE_204:		return 204;
    case DTU2XX_TXMODE_ADD16:	return 204;
    case DTU2XX_TXMODE_RAW:		return 188;
    case DTU2XX_TXMODE_MIN16:	return 204;
    default:					return 188;		// Should not occur
    }
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Dtu2xxBuf2RingBuffer -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Copies data from a contigious buffer to a ring buffer.
//
// Returns to number of copied bytes
//
//
Int Dtu2xxBuf2RingBuffer(
	UInt8* pBuf,						// buffer with data to copy
	Int NumToCopy,					// Number of bytes to copy
	PDTU2XX_RING_BUFFER pRingBuf)	// our ring buffer
{
	Int Copym_Count, BytesToWrap;

#ifdef DEBUG
	DTU2XX_LOG3( KERN_DEBUG, "Dtu2xxBuf2RingBuffer: NumToCopy = %d, pRingBuf->m_TotalSize"
		   " = %d, pRingBuf->m_Load = %d", NumToCopy, pRingBuf->m_TotalSize,
		   pRingBuf->m_Load );
#endif
	if ( NumToCopy == 0 ) {
		return 0;
	}

#ifdef DEBUG
	// Pre-conditions
	if ( !(	!((pRingBuf->m_Load >= 0) && (pRingBuf->m_Load <= pRingBuf->m_TotalSize)) ||
			!(pRingBuf->m_pWrPtr >= pRingBuf->m_pStart) ||
			!(pRingBuf->m_pWrPtr < (pRingBuf->m_pStart+pRingBuf->m_TotalSize)) ||
			!(pRingBuf->m_pRdPtr >= pRingBuf->m_pStart) ||
			!(pRingBuf->m_pRdPtr < (pRingBuf->m_pStart+pRingBuf->m_TotalSize)) ))
	{
		return -1;
	}
#endif
	
	// Check for overflow
	if ( NumToCopy > (pRingBuf->m_TotalSize - pRingBuf->m_Load) ) {
		Copym_Count = (pRingBuf->m_TotalSize - pRingBuf->m_Load);
	} else {
		Copym_Count = NumToCopy;
	}

	if ( Copym_Count == 0 ) {
		return 0;
	}

	// Check when the buffer wraps
	BytesToWrap = pRingBuf->m_TotalSize - (pRingBuf->m_pWrPtr - pRingBuf->m_pStart);

	if ( BytesToWrap >= Copym_Count ) {
		// No wrap problem => copy all in one go
		memcpy( pRingBuf->m_pWrPtr, pBuf, Copym_Count );

		// Update write pointer
		if ( Copym_Count == BytesToWrap ) {
			pRingBuf->m_pWrPtr = pRingBuf->m_pStart;
		} else {
			pRingBuf->m_pWrPtr += Copym_Count;
		}
	} else {
#ifdef DEBUG
	DTU2XX_LOG( KERN_DEBUG, "Dtu2xxBuf2RingBuffer Ring buffer wraps" );
#endif

        // First copy upto wrap point
		memcpy( pRingBuf->m_pWrPtr, pBuf, BytesToWrap );
		pRingBuf->m_pWrPtr = pRingBuf->m_pStart;
		// Copy remainder
		memcpy( pRingBuf->m_pWrPtr, &(pBuf[BytesToWrap]), (Copym_Count - BytesToWrap) );
		pRingBuf->m_pWrPtr += (Copym_Count - BytesToWrap);
	}
	// Update load
	pRingBuf->m_Load += Copym_Count;

#ifdef DEBUG
	// Post-conditions
	if ( !(	!((pRingBuf->m_Load >= 0) && (pRingBuf->m_Load <= pRingBuf->m_TotalSize)) ||
			!(pRingBuf->m_pWrPtr >= pRingBuf->m_pStart) ||
			!(pRingBuf->m_pWrPtr < (pRingBuf->m_pStart+pRingBuf->m_TotalSize)) ||
			!(pRingBuf->m_pRdPtr >= pRingBuf->m_pStart) ||
			!(pRingBuf->m_pRdPtr < (pRingBuf->m_pStart+pRingBuf->m_TotalSize)) ))
	{
		return -1;
	}
#endif

#ifdef DEBUG
	DTU2XX_LOG1( KERN_DEBUG, "Dtu2xxBuf2RingBuffer called; return Copym_Count = %d", Copym_Count );
#endif

	return Copym_Count; 
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Dtu2xxRingBuffer2Buf -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Copies data from a ring buffer to a contigious buffer
//
// Returns to number of copied bytes
//
Int Dtu2xxRingBuffer2Buf(
	PDTU2XX_RING_BUFFER pRingBuf,	// our ring buffer
	Int NumToCopy,					// Number of bytes to copy		
	UInt8* pBuf)						// buffer to receive the data
{
	Int Copym_Count, BytesToWrap;
#ifdef DEBUG
	DTU2XX_LOG3( KERN_DEBUG, "Dtu2xxRingBuffer2Buf: NumToCopy = %d, pRingBuf->m_TotalSize"
		   " = %d, pRingBuf->m_Load = %d", NumToCopy, pRingBuf->m_TotalSize,
		   pRingBuf->m_Load );
#endif
	if ( NumToCopy == 0 ) {
		return 0;
	}

#ifdef DEBUG
	// Pre-conditions
	if ( !( !((pRingBuf->m_Load >= 0) && (pRingBuf->m_Load <= pRingBuf->m_TotalSize)) ||
			!(pRingBuf->m_pWrPtr >= pRingBuf->m_pStart) ||
			!(pRingBuf->m_pWrPtr < (pRingBuf->m_pStart+pRingBuf->m_TotalSize)) ||
			!(pRingBuf->m_pRdPtr >= pRingBuf->m_pStart) ||
			!(pRingBuf->m_pRdPtr < (pRingBuf->m_pStart+pRingBuf->m_TotalSize)) ))
	{
	DT
		return -1;
	}
#endif
	
	// Check for underflow
	if ( NumToCopy > pRingBuf->m_Load ) {
		Copym_Count = pRingBuf->m_Load;
	} else {
		Copym_Count = NumToCopy;
	}

	if (Copym_Count == 0 ) {
		return 0;
	}

	// Check when the buffer wraps
	BytesToWrap = pRingBuf->m_TotalSize - (pRingBuf->m_pRdPtr - pRingBuf->m_pStart);

	if ( BytesToWrap >= Copym_Count ) {
		// No wrap problem => copy all in one go
		memcpy( pBuf, pRingBuf->m_pRdPtr, Copym_Count );

		// Update write pointer
		if ( Copym_Count == BytesToWrap ) {
			pRingBuf->m_pRdPtr = pRingBuf->m_pStart;
		} else {
			pRingBuf->m_pRdPtr += Copym_Count;
		}
	} else {
#ifdef DEBUG
	DTU2XX_LOG( KERN_DEBUG, "Dtu2xxRingBuffer2Buf Ring buffer wraps" );
#endif

        // First copy upto wrap point
		memcpy( pBuf, pRingBuf->m_pRdPtr, BytesToWrap );
		pRingBuf->m_pRdPtr = pRingBuf->m_pStart;
		// Copy remainder
		memcpy( &(pBuf[BytesToWrap]), pRingBuf->m_pRdPtr, (Copym_Count-BytesToWrap) );
		pRingBuf->m_pRdPtr += (Copym_Count-BytesToWrap);
	}
	// Update load
	pRingBuf->m_Load -= Copym_Count;

#ifdef DEBUG
	// Post-conditions
	if ( !( !((pRingBuf->m_Load >= 0) && (pRingBuf->m_Load <= pRingBuf->m_TotalSize)) ||
			!(pRingBuf->m_pWrPtr >= pRingBuf->m_pStart) ||
			!(pRingBuf->m_pWrPtr < (pRingBuf->m_pStart+pRingBuf->m_TotalSize)) ||
			!(pRingBuf->m_pRdPtr >= pRingBuf->m_pStart) ||
			!(pRingBuf->m_pRdPtr < (pRingBuf->m_pStart+pRingBuf->m_TotalSize)) ))
	{
		return -1;
	}
#endif
	
#ifdef DEBUG
	DTU2XX_LOG1( KERN_DEBUG, "Dtu2xxRingBuffer2Buf called; return Copym_Count = %d", Copym_Count );
#endif

	return Copym_Count; 
}
