//#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* Vpd.c *#*#*#*#*#*#*#*#*#*# (C) 2000-2006 DekTec
//
// This code is the intellectual property of DekTec Digital Video B.V. and may not be
// distributed, in its original or in altered form, without the explicit permission of
// DekTec Digital Video B.V.

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Change History -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//  2006.10.01	MG	Clean-up of code
//  2006.08.14	KvR	Changed to new interface, added support for DTU-245, several bugfixes
//	2004.07.28	MG	Created

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Include files -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
#include "Dtu2xx.h"
#include "Dtu2xxRegs.h"
#include "Dtu2xxVpd.h"
#include "EzUsb.h"

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Dtu2xxVpdRead -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Read an array of bytes from the serial EEPROM.
//
DTU2XX_RESULT Dtu2xxVpdRead(
    DTU2XX_FDO*  pFdo,	    // Device extension
    UInt8*	pBuf,			// Buffer for data read from serial EEPROM
    UInt  StartAddr,		// Start address (BYTE address)
    UInt  Length)			// Number of BYTES to read
{
	// Let's be robust
	if (Length == 0)
		return DTU2XX_OK;
	if ( StartAddr<0 || StartAddr>EEPROM_VPD_END )
		return DTU2XX_E_EEPROM_READ;
	if ( (StartAddr+Length) > EEPROM_VPD_END )
		return DTU2XX_E_EEPROM_READ;

	// Copy bytes from VPD cache
	memcpy( pBuf, &(pFdo->m_Vpd[StartAddr]), Length );

	return DTU2XX_OK;
}


//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Dtu2xxVpdWrite -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Write an array of bytes to the serial EEPROM.
//
DTU2XX_RESULT Dtu2xxVpdWrite(
    DTU2XX_FDO  * pFdo,	    // Device extension
    UInt8*  pBuf,			// Buffer with data to write to serial EEPROM
    UInt  StartAddr,		// Start address (BYTE address)
    UInt  Length)			// Number of BYTES to read
{
	Int ret=0;

	// Let's be robust
	if (Length == 0)
		return DTU2XX_OK;
	if ( StartAddr<0 || StartAddr>EEPROM_VPD_END )
		return DTU2XX_E_EEPROM_WRITE;
	if ( (StartAddr+Length) > EEPROM_VPD_END )
		return DTU2XX_E_EEPROM_WRITE;

	// First copy to EEPROM cache
	memcpy( &(pFdo->m_Vpd[StartAddr]), pBuf, Length );

	// Update actual EEPROM to macth our cache
	ret = Dtu2xxWriteEeProm( pFdo, StartAddr, Length, pBuf );
	if ( ret != 0 )
		return DTU2XX_E_EEPROM_WRITE;
	return DTU2XX_OK;
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Dtu2xxReadEeProm -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Gets the contents of the I2C EEPROM on the DTU-2xx
//
DTU2XX_RESULT Dtu2xxReadEeProm(
	DTU2XX_FDO* pFdo,			// Functional device object
	Int Offset,					// Offset address within EEPROM (i.e. start read here)
	Int NumBytes,				// Number of bytes to get (i.e. size of pEepromBuf)
	UInt8* pEepromBuf)			// Buffer that receives the EEPROM contents
{
	Int Status=0;
	Int TransferSize, ActualTransferred, BufIndex;
	Int BytesLeft = NumBytes;	// Bytes left to transfer

	// Nothing to transfer
	if ( BytesLeft == 0 )
		return DTU2XX_OK;

	BufIndex = 0;
    while ( BytesLeft > 0 )
	{
		// Do not transfer more than maximum allowed
		TransferSize = ( BytesLeft > EZUSB_EP0_MAX_TR_SIZE ) ? 
									EZUSB_EP0_MAX_TR_SIZE : BytesLeft;

		Status = Dtu2xxVendorRequest(
								pFdo, DTU2XX_USB_READ_EEPROM,
								Offset+BufIndex,	// Continue where we left of
								0,
								DTU2XX_DIRECTION_READ,
								TransferSize, &(pEepromBuf[BufIndex]),
								&ActualTransferred);

		// Transfer must be succesful
		if ( Status != 0 )
			return DTU2XX_E_EEPROM_READ;
		

		// Must have transferred all
		if ( ActualTransferred != TransferSize )
			return DTU2XX_E_EEPROM_READ;

		// update bytes left and index with buffer
		BytesLeft -= ActualTransferred;
		BufIndex += ActualTransferred;
	}
	return DTU2XX_OK;
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Dtu2xxWriteEeProm -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Sets the contents of the I2C EEPROM on the DTU-2xx
//
DTU2XX_RESULT Dtu2xxWriteEeProm(
	DTU2XX_FDO*  pFdo,		// Functional device object
	Int Offset,				// Offset address within EEPROM (i.e. start write here)
	Int NumBytes,			// Number of bytes to set (i.e. size of pEepromBuf)
	UInt8* pEepromBuf)		// Buffer with EEPROM contents to write
{
	Int Status=0;
	Int TransferSize, ActualTransferred, BufIndex;
	Int BytesLeft = NumBytes;	// Bytes left to transfer

	// Nothing to transfer
	if ( BytesLeft == 0 )
		return DTU2XX_OK;

	BufIndex = 0;
    while ( BytesLeft > 0 )
	{
		// Do not transfer more than maximum allowed
		TransferSize = ( BytesLeft > EZUSB_EP0_MAX_TR_SIZE ) ? 
									EZUSB_EP0_MAX_TR_SIZE : BytesLeft;

		Status = Dtu2xxVendorRequest(
								pFdo, DTU2XX_USB_WRITE_EEPROM,
								Offset+BufIndex,	// Continue where we left of
								0,
								DTU2XX_DIRECTION_WRITE,
								TransferSize, &(pEepromBuf[BufIndex]),
								&ActualTransferred);

		// Transfer must be succesful
		if ( Status != 0 )
			return DTU2XX_E_EEPROM_WRITE;
	        
		// Must have transferred all
		if ( ActualTransferred != TransferSize )
			return DTU2XX_E_EEPROM_WRITE;
			
		// update bytes left and index with buffer
		BytesLeft -= ActualTransferred;
		BufIndex += ActualTransferred;
	}
	return DTU2XX_OK;
}
